//Distributed under Kellen's sort-of-free license.
//http://people.colgate.edu/kmyers/license.html
//(c) Kellen Myers and Joseph Parrish 2007

// vim:set ts=4:

#include <stdlib.h>
#include <string.h>
#include "colorings.h"

static int iter_read(iterator_t *iter, coloring_t *c, unsigned char invert)
{
	int mask = 1 << (iter->abs_bit & 7);
	int loc = iter->abs_bit >> 3;
	// the !! operator maps all non-zero onto 1.
	return !!((c->bits[loc] ^ invert) & mask);
}

int iter_pos(iterator_t *iter)
{
	return iter->abs_bit + 1;
}

static void iter_overflow(iterator_t *iter)
{
	iter->overflow = 1;
	iter->abs_bit = 0;
}

static void iter_next(iterator_t *iter)
{
	if (iter->held) {
		iter->overflow = 1;
		return;
	}

	iter->abs_bit++;
	if (iter->abs_bit > iter->max_bit) {
		iter_overflow(iter);
	}
}

void iter_hold(iterator_t *iter)
{
	iter->abs_bit = iter->max_bit;
	iter->held = 1;
}

void iter_reset(iterator_t *iters, int num_vars, coloring_t *c)
{
	int x;
	for (x = 0; x < num_vars; x++) {
		iters[x].abs_bit = 0;
		iters[x].max_bit = c->last_bit - 1;
		iters[x].overflow = 0;
		iters[x].held = 0;
		while (iter_read(&iters[x], c, 0) != color_read(c, c->last_bit)) {
			iter_next(&iters[x]);
		}
	}
}

static void set_blue(coloring_t *c, int n)
{
	int loc;
	int mask;

	loc = (n - 1) >> 3;
	mask = 1 << ((n - 1) & 7);
	c->bits[loc] |= mask;
}

static void set_red(coloring_t *c, int n)
{
	int loc;
	int mask;

	loc = (n - 1) >> 3;
	mask = 1 << ((n - 1) & 7);
	c->bits[loc] &= ~mask;
}

static int spin(int n)
{
	double d = rand();
	d /= (double)RAND_MAX;
	d *= (double)n;
	return (int)d;
}

void color_randomly(coloring_t *c, int min_bit, int max_bit)
{
	int i;
	int need;

	c->max_bit = max_bit;
	c->last_bit = spin(max_bit - min_bit + 1) + min_bit;
	need = (c->last_bit + 7) >> 3;
	if (c->bits) free(c->bits);
	c->bits = (unsigned char *)malloc(need);

	set_red(c, 1);
	for (i = 2; i <= c->last_bit; i++) {
		if (rand() & 256) {
			set_blue(c, i);
		} else {
			set_red(c, i);
		}
	}
}

void color_restart(coloring_t *c)
{
	c->last_bit = 1;
	set_red(c, 1);
}

coloring_t *color_new(int n)
{
	coloring_t *c = (coloring_t *)malloc(sizeof(coloring_t));
	// round up to multiple of 4.
	n = (n + 3) & ~3;
	c->bits = (unsigned char *)malloc(n);
	c->max_bit = n << 3;
	return c;
}

void color_free(coloring_t *c)
{
	free(c->bits);
	free(c);
}

int color_read(coloring_t *c, int n)
{
	// the !! operator maps all non-zero onto 1.
	return !!(c->bits[(n - 1) >> 3] & (1 << ((n - 1) & 7)));
}

void color_grow(coloring_t *c)
{
	c->last_bit++;
	if (c->last_bit > c->max_bit) {
		c->max_bit *= 2;
		c->bits = (unsigned char *)realloc(c->bits, c->max_bit >> 3);
	}
	set_red(c, c->last_bit);
}

void color_backtrack(coloring_t *c)
{
	while (color_read(c, c->last_bit)) {
		c->last_bit--;
	}
	set_blue(c, c->last_bit);
}

void copy_coloring(coloring_t *dst, coloring_t *src)
{
	int need = (src->last_bit + 7) >> 3;

	if (dst->bits && (dst->max_bit < src->last_bit)) {
		free(dst->bits);
		dst->bits = NULL;
	}
	if (dst->bits == NULL) {
		dst->max_bit = need << 3;
		dst->bits = (unsigned char *)malloc(need);
	}
	dst->last_bit = src->last_bit;
	memcpy(dst->bits, src->bits, need);
}

int next(coloring_t *c, iterator_t *iters, int x, int num_vars)
{
	unsigned char invert;

	if (x == num_vars) {
		// done iterating.
		return 0;
	}

	if (color_read(c, c->last_bit)) invert = 0x00;
	else invert = 0xff;

	do {
		iter_next(&iters[x]);
		if (iters[x].overflow) {
			iters[x].overflow = 0;
			if (!next(c, iters, x + 1, num_vars))
				return 0;
		}
	} while (!iter_read(&iters[x], c, invert));
	return 1;
}

void fill_vars(iterator_t *iters, int *var_array, int num_vars)
{
	int i;
	for (i = 0; i < num_vars; i++) {
		var_array[i] = iter_pos(&iters[i]);
	}
}

